/*
 * Proiect PM
 */

#include "avr-wifi.h"
#include "lcd.h"
#include "uart.h"

#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>

#include <stdlib.h>
#include <string.h>

/* Tensiunea de referinta utilizata de ADC. */
#define ADC_AREF_VOLTAGE 5
#define ADC_CHANNEL 0
#define BTN_DEBOUNCE 10
#define UART_RX1_BUFFER_SIZE 1024
#define USART1_LARGE_BUFFER

volatile uint16_t ADC_value;
volatile uint16_t humidity;

static bool manual_mode = 0;
static uint16_t humidity_min = 10;
static uint16_t humidity_desired = 20;
static bool watering_in_progress = 0;
static bool watering_enabled = 0;
static int button_hold_count;

static uint8_t smile[] = {0b01010, 0b11011, 0b0000,  0b00100,
                          0b00000, 0b10001, 0b01110, 0b00000};

/*
// Wifi access point ussid and password (to connect to)
**/
static char USSID[] = "AndroidAP9862";
static char PASSWD[] = "shimotsuki";
static char URL[] = "192.168.43.200";
static int PORT = 5000;

/*
 * Functia initializeaza convertorul Analog-Digital.
 */
void ADC_init(void) {
  // enable ADC with:
  // * reference AVCC with external capacitor at AREF pin
  // * without left adjust of conversion result
  // * no auto-trigger
  // * interrupt
  // * prescaler at 32
  ADMUX = (1 << REFS0) | (1 << ADC_CHANNEL);
  ADCSRA = (1 << ADEN) | (1 << ADIE) | (5 << ADPS0);
}

void TIMER1_init() {
  // initialize timer to trigger COMPA every 100ms
  TCCR1B = (5 << CS10) | (1 << WGM12);
  TIMSK1 |= (1 << OCIE1A);
  OCR1A = 1561;
}

/*
 * Functia primeste o valoare numerica raw citita de convertul Analog-Digital
 * si calculeaza tensiunea (in volti) pe baza tensiunei de referinta.
 */
double ADC_voltage(int raw) { return raw * ADC_AREF_VOLTAGE / 1023.0; }

void Humidity_read() {
  humidity = (4.0 - ADC_voltage(ADC_value)) * 100 / 3;
  if (humidity < 0) {
    humidity = 0;
  }

  if (humidity > 100) {
    humidity = 0;
  }
}

/*
 * Functia afiseaza pe display-ul grafic valorile citite de la senzorul de
 * umiditate.
 */
void ADC_show(void) {
  char buf[32];

  // La ADC-ul microcontroller-ului avem conectat:
  //   ADC0: senzor umiditate
  sprintf(buf, "%s- %d%% ", "Humidity ", (int)humidity);
  LCD_printAt(0x00, buf);
}

void ADC_send() {
  wifi_connect(USSID, PASSWD);

  char buf[32];
  uart_flush();
  wifi_link_open("TCP", URL, PORT);
  sprintf(buf, "%d\r\n", (int)humidity);
  wifi_send(buf);
}

void CheckManualMode() {
  if ((PIND & (1 << PD6)) == 0) {
    button_hold_count++;
    if (button_hold_count == BTN_DEBOUNCE) {
      button_hold_count = 0;
      manual_mode = ~manual_mode;
      PORTD &= ~(1 << PD7);
    }
  }
}

void Try_water() {
  CheckManualMode();

  if (manual_mode) {
    LCD_printAt(0x28, "manual_mode      ");
    return;
  }

  watering_in_progress = false;
  PORTB &= ~(1 << PB0);

  if (!watering_enabled) {
    LCD_printAt(0x28, "  Just observing");
    LCD_putCharAt(0x40, 0);
  } else if (watering_in_progress) {
    LCD_printAt(0x28, "Water incoming   ");
    if (humidity <= humidity_desired) {
      PORTB |= (1 << PB0);
      watering_in_progress = true;
    }
  } else {
    if (humidity < humidity_min) {
      PORTB |= (1 << PB0);
      watering_in_progress = true;
      LCD_printAt(0x28, "Water incoming   ");
    } else {
      LCD_printAt(0x28, "Ready for action");
    }
  }
}

void IO_init() {
  DDRD |= (1 << PD7);

  DDRD &= ~(1 << PD6);
  PORTD |= (1 << PD6);

  // backlight LCD
  DDRC |= (1 << PC2);
  PORTC |= (1 << PC2);

  DDRB |= (1 << PB0);

  // Activez PIN CHANGE INTERRUPT ENABLE PCINT24:31 (portul D)
  PCICR |= (1 << PCIE3);
  // Activez intreruperea pe butoanele PD6
  PCMSK3 |= (1 << PCINT30);
}

/** Handle ESP8266 events */
void wifi_event_handler(int event, const char *input, const int len) {
  switch (event) {
  case WIFI_EVENT_ANY: {
    return;
  } break;

  case WIFI_EVENT_ERROR: {
    LCD_printAt(0x28, "ERROR\n  :<");
  } break;

  // ESP8266 is not responding
  case WIFI_EVENT_TIMEOUT: {
    LCD_printAt(0x28, "TIMEOUT\n  :<");
  } break;

  // Wifi was connected
  case WIFI_EVENT_WIFI_CONNECTED: {
    LCD_printAt(0x28, "Wifi connected\n  :)");
  } break;

  // Wifi was disconnected
  case WIFI_EVENT_WIFI_DISCONNECTED: {
    LCD_printAt(0x28, "Wifi disconnected\n  :(");
  } break;

  // Connect to host
  case WIFI_EVENT_CONNECTED: {
    LCD_printAt(0x28, "Fetch data...");
  } break;

  // Host sends data
  case WIFI_EVENT_DATA: {
    //
    // Look for JSON data in response
    //
  }
  }
}

ISR(ADC_vect) {
  // Intreruperea este triggered cand o conversie este finalizata. */
  ADC_value = ADC;
  ADMUX = (ADMUX & ~(0x1f << MUX0)) | ADC_CHANNEL;

  Humidity_read();
  ADC_show();
  Try_water();
  //ADC_send();
}

ISR(TIMER1_COMPA_vect) {
  // start an ADC conversion
  ADCSRA |= (1 << ADSC);
}

ISR(PCINT3_vect) {
  if ((PIND & (1 << PD6)) == 0) {
    PORTD ^= (1 << PD7);
    watering_enabled = !watering_enabled;
    button_hold_count = 0;

    if (manual_mode) {
      PORTB ^= (1 << PB0);
    }
  }
}

int main() {
  // Init UART
  uart_init(UART_BAUD_SELECT(2400, F_CPU));

  IO_init();
  LCD_init();
  ADC_init();
  TIMER1_init();
  //wifi_init();
  //wifi_connect(USSID, PASSWD);

  LCD_writeInstr(LCD_INSTR_CGRAM);
  for (uint8_t i = 0; i < sizeof(smile); ++i)
    LCD_writeData(smile[i]);
  LCD_writeInstr(LCD_INSTR_DDRAM);

  sei();

  for (;;)
    ;

  return 0;
}
